_ignore_statuses = {200, 204, 206, 409}


def handle_v3io_errors(response, file_path):
    if response.status_code in _ignore_statuses:
        return response.body
    if response.status_code == 404:
        raise FileNotFoundError(f'{file_path!r} not found')
    raise Exception(
        f'{response.status_code} received while accessing {file_path!r}')



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import datetime
from pathlib import Path

import dask.bag as db
import dask.dataframe as dd
import pytest
from conftest import test_container, test_dir

csv_data = b'''
name,item,price,quantity
Rick,Vodka,13.2,22
Jerry,Beer,34.2,300
Beth,Lettuce,1.2,13
Summer,M&M,2.2,7
Morty,Twix,1.7,5
'''


tests_root = Path(__file__).absolute().parent
test_pq = tests_root / 'sanchez.pq'
with test_pq.open('rb') as fp:
    parquet_data = fp.read()

data_by_type = {
    'csv': csv_data,
    'pq': parquet_data,
}


def test_text(tmp_obj):
    uri = f'v3io://{tmp_obj.path}'
    file = db.read_text(uri)
    data, = file.compute(scheduler='single-threaded')
    data = data.encode('utf-8')
    assert tmp_obj.data == data, 'bad data'


read_cases = [
    # read_fn, typ
    (dd.read_csv, 'csv'),
    (dd.read_parquet, 'pq'),
]


@pytest.mark.parametrize('read_fn, typ', read_cases)
def test_read(read_fn, typ, client, new_file):
    assert typ in data_by_type, f'unknown type: {typ!r}'
    data = data_by_type[typ]
    file_name = datetime.now().strftime('test-%Y%m%d%H%M.') + typ
    file_path = f'/{test_dir}/{file_name}'
    new_file(client, file_path, data)

    df = read_fn(f'v3io://{test_container}{file_path}')
    assert 4 == len(df.columns), '# of columns'
    assert 5 == len(df), '# of rows'



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime, timezone
from os.path import basename, dirname
from pathlib import Path

import fsspec
import pytest
from conftest import test_container, test_dir

from v3iofs import V3ioFS
from v3iofs.fs import parse_time
from v3iofs.path import split_container

path_types = [
    str,
    Path,
]


@pytest.mark.parametrize('path_cls', path_types)
def test_ls(fs: V3ioFS, new_file, path_cls):
    new_file(fs._client, f'{test_dir}/test-file')  # Make sure dir exists
    path = path_cls(f'/{test_container}/{test_dir}/')
    out = fs.ls(path)
    assert len(out) > 0, 'nothing found'
    assert all(isinstance(p, dict) for p in out), 'not dict'

    out = fs.ls(path, detail=False)
    assert len(out) > 0, 'nothing found'
    assert all(isinstance(p, str) for p in out), 'not string'


def test_rm(fs: V3ioFS, tmp_obj):
    path = tmp_obj.path
    fs.rm(path)
    out = fs.ls(dirname(path), detail=False)
    assert path not in out, 'not deleted'


def test_touch(fs: V3ioFS, tmp_obj):
    path = tmp_obj.path
    fs.touch(path)
    container, path = split_container(path)
    resp = fs._client.get_object(container, path)
    assert resp.body == b'', 'not truncated'


now = datetime(2020, 1, 2, 3, 4, 5, 6789, tzinfo=timezone.utc)
parse_time_cases = [
    # value, expected, raises
    ('', None, True),
    (now.strftime('%Y-%m-%d'), None, True),
    (now.strftime('%Y-%m-%dT%H:%M:%S.%f%z'), now.timestamp(), False),
    (now.strftime('%Y-%m-%dT%H:%M:%S.%fZ'), now.timestamp(), False),
]


@pytest.mark.parametrize('value, expected, raises', parse_time_cases)
def test_parse_time(value, expected, raises):
    if raises:
        with pytest.raises(ValueError):
            parse_time(value)
        return

    out = parse_time(value)
    assert expected == out


def load_tree(fs, root):
    out = {}
    for entry in fs.ls(root, detail=True):
        name = basename(entry['name'])
        if entry['type'] == 'file':
            with fs.open(entry['name'], 'rb') as fp:
                out[name] = fp.read()
        elif entry['type'] == 'directory':
            out[name] = load_tree(fs, f'{root}/{name}')
        else:
            raise TypeError(f'unknown entry type: {entry["type"]}')
    return out


def test_directory(fs, tree):
    out = load_tree(fs, tree.root)
    assert tree.data == out


def test_fsspec():
    fs = fsspec.filesystem("v3io")
    dirpath = f'/{test_container}/{test_dir}/fss'
    file_name = datetime.now().strftime('test_%f')
    filepath = f'{dirpath}/{file_name}.txt'
    with fs.open(filepath, 'wb') as fp:
        fp.write(b'123')
    for prefix in ['', 'v3io://']:
        files = fs.ls(prefix + dirpath, detail=True)
        assert len(files) == 1, \
            f'unexpected number of files {len(files)} in {prefix + dirpath}'
        assert fs.info(prefix + filepath)['type'] == 'file', \
            f'unexpected type in {prefix + dirpath}'
        assert fs.size(prefix + filepath) == 3, \
            f"unexpected file size in {prefix + dirpath}"
    with fs.open(prefix + filepath) as fp:
        data = fp.read()
    assert data == b'123', 'unexpected data'
    print(filepath)
    fs.rm(filepath)



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import datetime, timezone
from os import environ
from urllib.parse import urlparse

from fsspec.spec import AbstractFileSystem
from v3io.dataplane import Client
import v3io

from .file import V3ioFile
from .path import split_container, unslash, path_equal, strip_schema
from .utils import handle_v3io_errors

_file_key = 'key'
_dir_key = 'prefix'


class V3ioFS(AbstractFileSystem):
    """File system driver to v3io

    Parameters
    ----------
    v3io_api: str
        API host name (or V3IO_API environment)
    v3io_access_key: str
        v3io access key (or V3IO_ACCESS_KEY from environment)
    **kw:
        Passed to fsspec.AbstractFileSystem
    """

    protocol = 'v3io'

    def __init__(self, v3io_api=None, v3io_access_key=None, **kw):
        # TODO: Support storage options for creds (in kw)
        super().__init__(**kw)
        self._client = _new_client(v3io_api, v3io_access_key)

    def ls(self, path, detail=True, **kwargs):
        """Lists files & directories under path"""
        full_path = path
        path = strip_schema(path)
        container, path = split_container(path)
        if not container:
            return self._list_containers(detail)

        resp = self._client.get_container_contents(
            container=container,
            path=path,
            get_all_attributes=True,
            raise_for_status=v3io.dataplane.RaiseForStatus.never,
        )

        # Ignore 404's here
        if resp.status_code not in {200, 404}:
            raise Exception(
                f'{resp.status_code} received while accessing {path!r}')

        # If not data, try to find file in parent directory
        if not _has_data(resp):
            return [self._ls_file(container, path, detail)]

        out = (
                _resp_dirs(resp, container, detail) +
                _resp_files(resp, container, detail)
        )

        if not out:
            raise FileNotFoundError(f'{full_path!r} not found')

        return out

    def _ls_file(self, container, path, detail):
        # '/a/b/c' -> ('/a/b', 'c')
        dirname, _, filename = path.rpartition('/')
        resp = self._client.get_container_contents(
            container=container,
            path=dirname,
            get_all_attributes=True,
            raise_for_status=v3io.dataplane.RaiseForStatus.never,
        )

        full_path = f'/{container}/{path}'
        handle_v3io_errors(resp, full_path)
        contents = getattr(resp.output, 'contents', [])
        objs = [obj for obj in contents if path_equal(obj.key, path)]
        if not objs:
            raise FileNotFoundError(full_path)

        obj = objs[0]
        if not detail:
            return full_path
        return info_of(container, obj, _file_key)

    def _list_containers(self, detail):
        resp = self._client.get_containers(
            raise_for_status=v3io.dataplane.RaiseForStatus.never)
        handle_v3io_errors(resp, 'containers')
        fn = container_info if detail else container_path
        return [fn(c) for c in resp.output.containers]

    def copy(self, path1, path2, **kwargs):
        ...  # FIXME

    def _rm(self, path):
        container, path = split_container(path)
        if not container:
            raise ValueError(f'bad path: {path:r}')

        resp = self._client.delete_object(
            container=container,
            path=path,
            raise_for_status=v3io.dataplane.RaiseForStatus.never,
        )

        # Ignore 404's and 409's in delete
        if resp.status_code not in {200, 204, 404, 409}:
            raise Exception(
                f'{resp.status_code} received while accessing {path!r}')

    def touch(self, path, truncate=True, **kwargs):
        if not truncate:  # TODO
            raise ValueError('only truncate touch supported')

        path = strip_schema(path)
        container, path = split_container(path)
        resp = self._client.put_object(
            container, path,
            raise_for_status=v3io.dataplane.RaiseForStatus.never
        )

        handle_v3io_errors(resp, path)

    def info(self, path, **kw):
        """Details of entry at path

        Returns a single dictionary, with exactly the same information as
        ``ls`` would with ``detail=True``.

        Parameters
        ----------
        path: str
            Path to get info for
        **kw:
            Keyword arguments passed to `ls`

        Returns
        -------
        dict
            keys: name (full path in the FS), size (in bytes), type (file,
            directory, or something else) and other FS-specific keys.
        """

        path = strip_schema(path)
        out = self.ls(path, detail=True, **kw)
        entries = [o for o in out if path_equal(o['name'], path)]

        if len(entries) == 1:
            entry = entries[0]
            entry.setdefault('size', None)
            return entry

        if len(entries) >= 1 or out:
            return {'name': path, 'size': 0, 'type': 'directory'}

        raise FileNotFoundError(path)

    def _open(
            self,
            path,
            mode='rb',
            block_size=None,
            autocommit=True,
            cache_type='readahead',
            cache_options=None,
            **kw,
    ):
        return V3ioFile(
            fs=self,
            path=path,
            mode=mode,
            block_size=block_size,
            autocommit=autocommit,
            cache_type=cache_type,
            cache_options=cache_options,
            **kw,
        )


def container_path(container):
    return f'/{container.name}'


def parse_time(creation_date):
    # '2020-03-26T09:42:57.504000+00:00'
    # '2020-03-26T09:42:57.71Z'
    i = creation_date.rfind('+')  # If not found will be -1, good for Z
    dt = datetime.strptime(creation_date[:i], '%Y-%m-%dT%H:%M:%S.%f')
    dt = dt.replace(tzinfo=timezone.utc)
    return dt.timestamp()


def container_info(container):
    return {
        'name': container.name,
        'size': None,
        'created': parse_time(container.creation_date),
    }


_missing = object()
_extra_obj_attrs = [
    # dest, src, convert
    ('created', 'creating_time', parse_time),
    ('atime', 'access_time', parse_time),
    ('mode', 'mode', lambda v: int(v[1:], base=8)),  # '040755'
    ('gid', 'gid', lambda v: int(v, 16)),
    ('uid', 'uid', lambda v: int(v, 16)),
]


def obj_path(container, obj, name_key):
    path = unslash(getattr(obj, name_key))
    return f'/{container}/{path}'


def info_of(container_name, obj, name_key):
    info = {
        'name': obj_path(container_name, obj, name_key),
        'type': 'file' if hasattr(obj, 'size') else 'directory',
        'size': getattr(obj, 'size', 0),
        'mtime': parse_time(obj.last_modified),
    }

    for src, dest, conv in _extra_obj_attrs:
        val = getattr(obj, src, _missing)
        if val is _missing:
            continue
        info[dest] = conv(val)

    return info


def split_auth(url):
    """
    >>> split_auth('v3io://api_key:s3cr3t@domain.company.com')
    ('v3io://domain.company.com', 's3cr3t')
    >>> split_auth('v3io://domain.company.com')
    ('v3io://domain.company.com', '')
    """
    u = urlparse(url)
    if '@' not in u.netloc:
        return (url, '')

    auth, netloc = u.netloc.split('@', 1)
    if ':' not in auth:
        raise ValueError('missing : in auth')
    _, key = auth.split(':', 1)
    u = u._replace(netloc=netloc)
    return (u.geturl(), key)


def _resp_dirs(resp, container, detail):
    if not hasattr(resp.output, 'common_prefixes'):
        return []

    objs = resp.output.common_prefixes
    if not detail:
        return [obj_path(container, obj, _dir_key) for obj in objs]

    return [info_of(container, obj, _dir_key) for obj in objs]


def _resp_files(resp, container, detail):
    if not hasattr(resp.output, 'contents'):
        return []

    objs = resp.output.contents
    if not detail:
        return [obj_path(container, obj, _file_key) for obj in objs]

    return [info_of(container, obj, _file_key) for obj in objs]


def _has_data(resp):
    out = resp.output
    return hasattr(out, 'common_prefixes') or hasattr(out, 'contents')


def _new_client(v3io_api=None, v3io_access_key=None) -> Client:
    v3io_api = v3io_api or environ.get('V3IO_API')
    v3io_access_key = \
        v3io_access_key or environ.get('V3IO_ACCESS_KEY')

    return Client(
        endpoint=v3io_api,
        access_key=v3io_access_key,
    )



# Copyright 2020 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from subprocess import run
from sys import executable
from tempfile import mkdtemp

import v3iofs


def test_install():
    out = run([executable, 'setup.py', 'sdist'])
    assert out.returncode == 0, 'sdist'

    tmp_dir = mkdtemp(prefix='v3iofs-test')
    out = run([executable, '-m', 'venv', tmp_dir])
    assert out.returncode == 0, 'create venv'

    dist = f'./dist/v3iofs-{v3iofs.__version__}.tar.gz'
    py = f'{tmp_dir}/bin/python'

    out = run([py, '-m', 'pip', 'install', dist])
    assert out.returncode == 0, 'install'

    out = run([py, '-c', 'import v3iofs'])
    assert out.returncode == 0, 'import'
